/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: adi2recv.c,v 1.3 1998/02/17 20:44:15 gropp Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpid.h"
#include "mpiddev.h"
#include "reqalloc.h"
#include "queue.h"
#include "chgamma.h"

/* 
  This file contains the routines to handle receiving a message
 */

/***************************************************************************/
/*
 * Despite the apparent symmetry, receives are fundamentally different from
 * sends.  All receives happen by processing an incoming item of information
 * and checking it against known receives.
 *
 * Eventually, we may want to make RecvContig a special case (as in the
 * first generation ADI) to avoid the routine calls.
 */
/***************************************************************************/
/* Does this need to return msgrep if heterogeneous? */

/* 
 * Error handling needs additional work.  Is a request that has detected an
 * error complete?  Who sets it?  What if the error is ERR_TRUNCATE (where
 * we'd like to make everything work to the limit of the buffer)?
 */


void MPID_RecvContig(
	struct MPIR_COMMUNICATOR *comm_ptr, void *buf, int maxlen,
	int src_lrank, int tag, int context_id, MPI_Status *status,
	int *error_code
) {
	MPIR_RHANDLE rhandle;
	MPI_Request  request = (MPI_Request)&rhandle;

	DEBUG_INIT_STRUCT(request,sizeof(rhandle));
	MPID_RecvInit( &rhandle );

	*error_code = 0;
	MPID_IrecvContig(
		comm_ptr, buf, maxlen, src_lrank, tag, context_id, request,
		error_code
	);

	if (!*error_code)
		MPID_RecvComplete( request, status, error_code );
}


void MPID_IrecvContig(
	struct MPIR_COMMUNICATOR *comm_ptr, void *buf, int maxlen,
	int src_lrank, int tag, int context_id, MPI_Request request,
	int *error_code
) {

MPIR_RHANDLE *dmpi_unexpected, *rhandle = &request->rhandle;
int position;
MPID_Device *dev = MPID_devset->dev_list;

DEBUG_PRINT_ARGS("R starting IrecvContig");

/* The one error test that makes sense here */

if (buf != 0 || maxlen == 0) {

	/* 
	At this time, we check to see if the message has already been received.
	Note that we cannot have any thread receiving a message while 
	checking the queues.   In case we do enqueue the message, we set
	the fields that will need to be valid BEFORE calling this routine
	(this is extra overhead ONLY in the case that the message was
	unexpected, which is already the higher-overhead case).
	*/
	rhandle->is_complete	= 0;
	rhandle->contextid	= context_id;
	rhandle->buf		= buf;
	rhandle->len		= maxlen;
	rhandle->finish		= 0;

	/*
	Here we search for a matching recv without extracting from the
	unexpected queue (flag=0), if found, because it might be still
	incomplete (receiver handler may have run and check_device not yet).
	*/
	MPID_Atomic_Search_unexpected_queue_and_post(
		src_lrank, tag, context_id, 0, buf, maxlen,
		rhandle, &dmpi_unexpected, &position
	);

	if (dmpi_unexpected != NULL) {
		/*  Message has unexpectedly arrived.  */

		if ((void *)dmpi_unexpected == (void *)(-1)) {
			DEBUG_PRINT_MSG("R Unexpected but not yet enqueued");
#if AG_DEBUG
			printf("[%d] adi2recv: unexpd msg not yet enqueued\n",
				gamma_my_node());
			fflush(stdout);
#endif
			/*
			Go extract messages from the device to complete this
			entry of unexpected queue.
			*/
			do
			*error_code = (*dev->check_device)(dev,MPID_BLOCKING);
			while (!(MPID_Unexpected_entry_is_complete(position)));
			/*
			Now that the entry is complete, it can be extracted
			from unexpected queue.
			*/
			MPID_Atomic_Extract_unexpected_entry(
				position,&dmpi_unexpected
			);
			DEBUG_TEST_FCN(dmpi_unexpected->push,"req->push");
			*error_code = (*dmpi_unexpected->push)(
				rhandle, dmpi_unexpected
			);
		}
		else {
			DEBUG_PRINT_MSG("R Found in unexpected queue");
#if AG_DEBUG
			printf("[%d] adi2recv: unexpd msg already enqueued\n",
				gamma_my_node());
			fflush(stdout);
#endif
			MPID_Atomic_Extract_unexpected_entry(
				position,&dmpi_unexpected
			);
			DEBUG_TEST_FCN(dmpi_unexpected->push,"req->push");
			*error_code = (*dmpi_unexpected->push)(
				rhandle, dmpi_unexpected
			);
		}
		DEBUG_PRINT_MSG("R Exiting IrecvContig");
		/* This may or may not complete the message */
		return;
	}

	/* If we got here, the message is not yet available */
	/*    MPID_DRAIN_INCOMING */

	DEBUG_PRINT_MSG("R Exiting IrecvContig");
#if AG_DEBUG
	printf("[%d] adi2recv: posted a request\n",gamma_my_node());
	fflush(stdout);
#endif
	return;

}

*error_code = MPI_ERR_BUFFER;
return;

}


int MPID_RecvIcomplete(
	MPI_Request request, MPI_Status *status, int *error_code
) {
	MPIR_RHANDLE *rhandle = &request->rhandle;
	MPID_Device *dev;
	int lerr;

	*error_code = MPI_SUCCESS;

	if (!rhandle->is_complete) {
		dev = MPID_devset->dev_list;
		lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
		if (lerr < 0) {
			return 0;
		}
		if (lerr > 0 && rhandle->s.MPI_ERROR) {
			*error_code = lerr;
			return 0;
		}
		if (!rhandle->is_complete) {
			return 0;
		}
	}

	if (status) *status = rhandle->s;
	*error_code = rhandle->s.MPI_ERROR;
	return 1;
}


void MPID_RecvComplete(
	MPI_Request request, MPI_Status *status, int *error_code
) {
	MPIR_RHANDLE *rhandle = &request->rhandle;
	MPID_Device *dev;
	int lerr;

	dev = MPID_devset->dev_list;

	while (!rhandle->is_complete) {
		lerr = (*dev->check_device)( dev, MPID_BLOCKING );
		/*
		An error return from check_device might apply
		to a different request.  We break only if
		the request that encountered the error is the one
		we detected
		*/
		if (lerr > 0 && rhandle->s.MPI_ERROR) {
			*error_code = lerr;
			break;
		}
	}
	if (status) *status = rhandle->s;
	*error_code = rhandle->s.MPI_ERROR;
	return;
}
