/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: queue.c,v 1.2 1998/01/29 14:25:46 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */


/*#define DEBUG(a) {a;}*/
#define DEBUG(a)

/*
 * Routines for managing message queues in the device implementation.
 * Some devices provide their own queue management routines, and
 * do not need these.
 */

#include "mpid.h"
#include <stdio.h>
#include "queue.h"
#include "chgamma.h"

MPID_QHDR MPID_recvs;

/*
   Queueing unexpected messages and searching for them requires particular
   care because of the three cases:
   tag              source            Ordering
   xxx              xxx               Earliest message in delivery order
   MPI_ANY_TAG      xxx               Earliest with given source, not 
                                      lowest in tag value
   xxx              MPI_ANY_SOURCE    Earliest with given tag, not lowest
                                      in source.

   Note that only the first case is explicitly required by the MPI spec.
   However, there is a requirement on progress that requires SOME mechanism;
   the one here is easy to articulate and matches what many users expect.
   
   There are many comprimises that must be considered when deciding how to
   provide for these different criteria.  

   The code here takes the simple view that the number of unexpected messages
   is likely to be small, so it just keeps things in arrival order and
   searches for the first matching message. It might be nicer to have separate
   queues assocaited with each communicator. Hashing based on tag/source would
   also be possible, however it is unlikely to improve on the simple queue
   because the simple queue is still required, and the hash always adds
   overhead while only saving under some circumstances.

   The current system does a linear search through the entire list, and 
   thus will always give the earliest in delivery order AS RECEIVED BY THE
   ADI.  We've had trouble with message-passing systems that the ADI is
   using not providing a fair delivers (starving some sources); this 
   should be fixed by the ADI or underlying message-passing system rather
   than by this level of MPI routine.  
 */

void MPID_Dump_queues()
{
    MPID_Dump_queue( &MPID_recvs );
}
void mpid_dump_queues__()
{
    MPID_Dump_queue( &MPID_recvs );
}

void MPID_Dump_queue(MPID_QHDR *header)
{
	MPID_QEL *p;
	unsigned long i;

	if (!header) return;

	printf("[%d] Unexpected queue:\n", MPID_MyWorldRank);
	printf("[%d] there are %ld elms, head %ld, tail %ld\n",
		MPID_MyWorldRank,
		header->unexpected.how_many,
		header->unexpected.head,
		header->unexpected.tail
	);
	if (header->unexpected.how_many) {
	i=header->unexpected.head;
	for (;;) {
		p = &(header->unexpected.queue[i]);
		printf(
"[%d] pos = %ld: free = %d, next = %d, context_id = %d, tag = %d(%x), src = %d(%x), rndv_pending = %d, buf = %8.8lx, len = %d, rhandle %8.8lx\n",
MPID_MyWorldRank, i, p->free, p->next,
p->context_id, p->tag, p->tagmask, p->lsrc, p->srcmask,
p->rndv_pending, (unsigned long)(p->buf), p->len, (unsigned long)(p->ptr)
		);
		if (i == header->unexpected.tail) break;
		if (i == (unsigned long)p->next) {
printf("[%d] queue is inconsistent!!\n",MPID_MyWorldRank);
			break;
		}
		i = p->next;
	}
	}

	printf("[%d] Posted receive queue:\n", MPID_MyWorldRank );
	printf("[%d] there are %ld elms, head %ld, tail %ld\n",
		MPID_MyWorldRank,
		header->posted.how_many,
		header->posted.head,
		header->posted.tail
	);
	if (header->posted.how_many) {
	i=header->posted.head;
	for (;;) {
		p = &(header->posted.queue[i]);
		printf(
"[%d] pos = %ld: free = %d, next = %d, context_id = %d, tag = %d(%x), src = %d(%x), rndv_pending = %d, buf = %8.8lx, len = %d, rhandle %8.8lx\n",
MPID_MyWorldRank, i, p->free, p->next,
p->context_id, p->tag, p->tagmask, p->lsrc, p->srcmask,
p->rndv_pending, (unsigned long)(p->buf), p->len, (unsigned long)(p->ptr)
		);
		if (i == header->posted.tail) break;
		if (i == (unsigned long)p->next) {
fprintf(stdout,"[%d] queue is inconsistent!!\n",MPID_MyWorldRank);
			break;
		}
		i = p->next;
	}
	}
	fflush(stdout);
}


static int MPID_Enqueue(
	MPID_QUEUE *queue, int src_lrank, int tag, int context_id,
	MPIR_RHANDLE *rhandle, void *buf, int len
)
{
	MPID_QEL *p;
	unsigned long i;

	p = NULL;
	i = queue->tail;
	do {
		if (queue->queue[i].free) {
			p = &(queue->queue[i]);
			break;
		}
		i = (i+1) % AG_MAXQLEN;
	} while (i != queue->tail);
	if (p) {
		if (queue->how_many) {
			queue->queue[queue->tail].next = i;
			queue->tail = i;
		}
		queue->how_many++;
		p->free = 0;
		p->next = 0;
		/* Store the selection criteria is an easy-to-get place */
		p->context_id = context_id;
		if (tag == MPI_ANY_TAG) {
			p->tagmask = 0;       /* No need to store the tag */
		}
		else {
			p->tag     = tag; 
			p->tagmask = ~0;
		}
		if (src_lrank == MPI_ANY_SOURCE) {
			p->srcmask = 0;       /* No need to store the source */
		}
		else {
			p->lsrc    = src_lrank;
			p->srcmask = ~0;
		}
		p->rndv_pending = 0;
		p->buf = buf;
		p->len = len;
		p->ptr = rhandle;
		return MPI_SUCCESS;
	}
	return MPI_ERR_EXHAUSTED;
}


static int MPID_Enqueue_rndv(
	MPID_QUEUE *queue, int src_lrank, int tag, int context_id,
	MPIR_RHANDLE *rhandle, void *buf, int len
)
{
	MPID_QEL *p;
	unsigned long i;

	p = NULL;
	i = queue->tail;
	do {
		if (queue->queue[i].free) {
			p = &(queue->queue[i]);
			break;
		}
		i = (i+1) % AG_MAXQLEN;
	} while (i != queue->tail);
	if (p) {
		/*  Append new element at queue tail (non-empty queue).  */
		if (queue->how_many) {
			queue->queue[queue->tail].next = i;
			queue->tail = i;
		}
		queue->how_many++;
		p->free = 0;
		p->next = 0;
		/* Store the selection criteria is an easy-to-get place */
		p->context_id = context_id;
		if (tag == MPI_ANY_TAG) {
			p->tagmask = 0;       /* No need to store the tag */
		}
		else {
			p->tag     = tag; 
			p->tagmask = ~0;
		}
		if (src_lrank == MPI_ANY_SOURCE) {
			p->srcmask = 0;       /* No need to store the source */
		}
		else {
			p->lsrc    = src_lrank;
			p->srcmask = ~0;
		}
		p->rndv_pending = 1;
		p->buf = buf;
		p->len = len;
		p->ptr = rhandle;
		return MPI_SUCCESS;
	}
	return MPI_ERR_EXHAUSTED;
}


/*
 * Remove the given request from the queue (for MPID_Cancel)
 *
 * Returns 0 on success, MPI_ERR_INTERN on not found.
 */

int MPID_Dequeue(MPID_QUEUE *queue, MPIR_RHANDLE *rhandle)
{
	MPID_QEL *p;
	unsigned long i;
	int prev;

	/* Look for the one we need */
	if (queue->how_many) {
		p = NULL;
		i = queue->head;
		prev = i;
		for(;;) {
			if (queue->queue[i].ptr == rhandle) {
				p = &(queue->queue[i]);
				break;
			}
			if (i == queue->tail)
				break;
			prev = i;
			i = queue->queue[i].next;
		}
		if (p) {
			/*  Take element "p" (index "i") out.  */
			p->free = 1;
			queue->how_many--;
			if (!(queue->how_many)) {
				queue->head = 0;
				queue->tail = 0;
				return MPI_SUCCESS;
			}
			queue->queue[prev].next = p->next;
			if (i == queue->head)
				queue->head = p->next;
			if (i == queue->tail)
				queue->tail = prev;
			return MPI_SUCCESS;
		}
	}
	return MPI_ERR_INTERN;	/* It's not there. Oops */
}


/* 
   Search posted_receive queue for message matching criteria.
   A flag of 1 causes a successful search to delete the element found.
   handleptr is set to non-null if an entry is found.
 */

static int MPID_Search_posted_queue(
	int src, int tag, int context_id, int flag, MPIR_RHANDLE **handleptr
)
{
	MPID_QUEUE *queue = &MPID_recvs.posted;
	MPID_QEL *p;
	unsigned long i;
	int prev;

	DEBUG(printf( "[%d] Search posted: begin\n", MPID_MyWorldRank ));
	DEBUG(MPID_Dump_queue(&MPID_recvs));

	/*
	 * Look for the one we need.
	 * Note that by doing the tests using 'xor' and 'and' we don't need to
	 * worry about the value in tag when tagmask is zero. (i.e. a wildcard
	 * receive). This saves us from ever bothering to store a tag in that
	 * case.
	 * Remember that xor (^) is a bit wise not equal function.
	 */

	if (queue->how_many) {

	i = queue->head;
	prev = i;
	for (;;) {
		p = &(queue->queue[i]);
		if (
			context_id == p->context_id &&
			((tag ^ p->tag) & p->tagmask) == 0 &&
			((src ^ p->lsrc)& p->srcmask) == 0
		) {
			DEBUG(printf( "[%d] Search posted: found\n",
				MPID_MyWorldRank ));
			*handleptr = p->ptr;
			if (flag) {
				/*  Take element "p" (index "i") out.  */
				DEBUG(printf( "[%d] Search posted: remove\n",
					MPID_MyWorldRank ));
				p->free = 1;
				queue->how_many--;
				if (!(queue->how_many)) {
					queue->head = 0;
					queue->tail = 0;
				}
				else {
					queue->queue[prev].next = p->next;
					if (i == (unsigned long)queue->head)
						queue->head = p->next;
					if (i == (unsigned long)queue->tail)
						queue->tail = prev;
				}
			}
			DEBUG(MPID_Dump_queue(&MPID_recvs));
			return MPI_SUCCESS;
		}
		if (i == (unsigned long)queue->tail) break;
		prev = i;
		i = p->next;
	}
	*handleptr = NULL;
	DEBUG(printf( "[%d] Search posted: not found\n", MPID_MyWorldRank ));
	DEBUG(MPID_Dump_queue(&MPID_recvs));
	return MPI_SUCCESS;

	}

	*handleptr = NULL;
	DEBUG(printf( "[%d] Search posted: not found\n",
		MPID_MyWorldRank ));
	DEBUG(MPID_Dump_queue(&MPID_recvs));
	return MPI_SUCCESS;
}


/*
Search unexpected_recv queue for message matching criteria.
A flag of 1 causes a successful search to delete the element found.
"handleptr" is non-null if an element is found.  In this case, and only
in this case, "position" indicates the selected entry in unexpected queue.
*/
static int MPID_Search_unexpected_queue(
	int src, int tag, int context_id, int flag,
	MPIR_RHANDLE **handleptr, int *position
)
{
	MPID_QUEUE    *queue = &MPID_recvs.unexpected;
	MPID_QEL      *p;
	int           tagmask, srcmask;
	unsigned long i;
	int           prev;

	DEBUG(printf( "[%d] searching for (%d,%d,%d) in unexpected queue\n", 
		MPID_MyWorldRank, context_id, tag, src )
	);
	/* Look for the one we need */
	if (!queue->how_many) {
		*handleptr = 0;
		return MPI_SUCCESS;
	}
	tagmask = (tag == MPI_ANY_TAG)    ? 0 : ~0;
	srcmask = (src == MPI_ANY_SOURCE) ? 0 : ~0;
	i=queue->head;
	prev = i;
	for (;;) {
		p = &(queue->queue[i]);
DEBUG(printf("[%d] in unexpected list, looking at (%d,%d,%d) at %lx\n",
MPID_MyWorldRank, p->context_id, (p->tag & tagmask), (p->lsrc & srcmask), 
(MPI_Aint)p));
		if (
			context_id == p->context_id &&
			(((tag ^ p->tag) & tagmask) == 0) &&
			(((src ^ p->lsrc) & srcmask) == 0)
		) {
			DEBUG(printf( "[%d] Search unxp: found\n",
				MPID_MyWorldRank ));
			*handleptr = p->ptr;
			*position = i;
			if (flag) {
				/*  Take element "p" (index "i") out.  */
				DEBUG(printf( "[%d] Search unxp: remove\n",
					MPID_MyWorldRank ));
				p->free = 1;
				queue->how_many--;
				if (!(queue->how_many)) {
					queue->head = 0;
					queue->tail = 0;
				}
				else {
					queue->queue[prev].next = p->next;
					if (i == (unsigned long)queue->head)
						queue->head = p->next;
					if (i == (unsigned long)queue->tail)
						queue->tail = prev;
				}
			}
			DEBUG(MPID_Dump_queue(&MPID_recvs));
			return MPI_SUCCESS;
		}
		if (i == queue->tail) break;
		prev = i;
		i = p->next;
	}
	*handleptr = 0;
	return MPI_SUCCESS;
}


int MPID_Unexpected_entry_is_complete(int position)
{
	return (
		(void *)MPID_recvs.unexpected.queue[position].ptr
		!= (void *)(-1)
	);
}


MPIR_RHANDLE *MPID_Get_unexpected_entry(int position)
{
	return MPID_recvs.unexpected.queue[position].ptr;
}


/*
 *  Thread-safe and handler-safe routines.
 */

int		SUQP_src_lrank, SUQP_tag, SUQP_context_id, SUQP_flag,
		SUQP_position, SUQP_len, SUQP_i, SUQP_prev;
void		*SUQP_buf;
MPIR_RHANDLE	*SUQP_request, *SUQP_rhandle;
MPID_QEL	*SUQP_p;


static void do_MPID_Atomic_Search_unexpected_queue(void)
{
	MPID_Search_unexpected_queue(
		SUQP_src_lrank,SUQP_tag,SUQP_context_id,SUQP_flag,
		&SUQP_rhandle,&SUQP_position
	);
}

void MPID_Atomic_Search_unexpected_queue(
	int src, int tag, int context_id, int flag,
	MPIR_RHANDLE **rhandleptr, int *position
)
{
	SUQP_src_lrank = src;
	SUQP_tag = tag;
	SUQP_context_id = context_id;
	SUQP_flag = flag;
	gamma_atomic(do_MPID_Atomic_Search_unexpected_queue);
	*rhandleptr = SUQP_rhandle;
	if (SUQP_rhandle) *position = SUQP_position;
}


/*
 * The behaviour of MPID_Atomic_Search_unexpected_queue_and_post is as follows:
 * first check the unexpected queue, then post the receive if a matching
 * receive is not found in the unexpected queue.
 * "handleptr" is non-null if an element is found.
 * In this case, and only in this case, upon return "position" indicates
 * the selected entry in unexpected queue.
 * The entry is removed from the unexpected queue if it is found and
 * flag == 1.
 */
static void do_MPID_Atomic_Search_unexpected_queue_and_post(void)
{
	/*  Search a matching receive in unexpected queue.  */
	MPID_Search_unexpected_queue(
		SUQP_src_lrank,SUQP_tag,SUQP_context_id,SUQP_flag,
		&SUQP_rhandle,&SUQP_position
	);
	if (!SUQP_rhandle) {
		/*  Not found.  Add to the posted receive queue.  */
		MPID_Enqueue(
			&MPID_recvs.posted,SUQP_src_lrank,SUQP_tag,
			SUQP_context_id,SUQP_request,SUQP_buf,SUQP_len
		);
	}
	return;
}

void MPID_Atomic_Search_unexpected_queue_and_post(
	int src_lrank, int tag, int context_id, int flag, void *buf, int len,
	MPIR_RHANDLE *request, MPIR_RHANDLE **rhandleptr, int *position
)
{
	SUQP_src_lrank = src_lrank;
	SUQP_tag = tag;
	SUQP_context_id = context_id;
	SUQP_flag = flag;
	SUQP_buf = buf;
	SUQP_len = len;
	SUQP_request = request;
	SUQP_rhandle = NULL;
	/*  Search a matching receive in unexpected queue.  */
	gamma_atomic(do_MPID_Atomic_Search_unexpected_queue_and_post);
	*rhandleptr = SUQP_rhandle;
	if (SUQP_rhandle) *position = SUQP_position;
}


static void do_MPID_Atomic_Search_posted_queue(void)
{
	MPID_Search_posted_queue(
		SUQP_src_lrank,SUQP_tag,SUQP_context_id,SUQP_flag,
		&SUQP_rhandle
	);
}

void MPID_Atomic_Search_posted_queue(
	int src, int tag, int context_id, int flag,
	MPIR_RHANDLE **rhandleptr
)
{
	SUQP_src_lrank = src;
	SUQP_tag = tag;
	SUQP_context_id = context_id;
	SUQP_flag = flag;
	gamma_atomic(do_MPID_Atomic_Search_posted_queue);
	*rhandleptr = SUQP_rhandle;
}


static void do_MPID_Atomic_Post_rndv(void)
{
	MPID_Enqueue_rndv(
		&MPID_recvs.posted,SUQP_src_lrank,SUQP_tag,
		SUQP_context_id,SUQP_request,SUQP_buf,SUQP_len
	);
}

void MPID_Atomic_Post_rndv(
	int src_lrank, int tag, int context_id, void *buf, int len,
	MPIR_RHANDLE *request
)
{
	SUQP_src_lrank = src_lrank;
	SUQP_tag = tag;
	SUQP_context_id = context_id;
	SUQP_buf = buf;
	SUQP_len = len;
	SUQP_request = request;
	gamma_atomic(do_MPID_Atomic_Post_rndv);
}


static void do_MPID_Atomic_Unexp(void)
{
	MPID_Enqueue(
		&MPID_recvs.unexpected,SUQP_src_lrank,SUQP_tag,
		SUQP_context_id,SUQP_request,SUQP_buf,SUQP_len
	);
}

void MPID_Atomic_Unexp(
	int src_lrank, int tag, int context_id, void *buf, int len,
	MPIR_RHANDLE *request
)
{
	SUQP_src_lrank = src_lrank;
	SUQP_tag = tag;
	SUQP_context_id = context_id;
	SUQP_buf = buf;
	SUQP_len = len;
	SUQP_request = request;
	gamma_atomic(do_MPID_Atomic_Unexp);
}


static void do_MPID_Atomic_Extract_unexpected_entry(void)
{
	/*  Take entry "position" out from unexpected queue.  */
	/*  Similar to MPID_Dequeue().  */

	if (MPID_recvs.unexpected.how_many) {
		SUQP_p = NULL;
		SUQP_i = MPID_recvs.unexpected.head;
		SUQP_prev = SUQP_i;
		for (;;) {
			if (SUQP_i == SUQP_position) {
				SUQP_p = &(MPID_recvs.unexpected.queue[SUQP_i]);
				break;
			}
			if ((unsigned long)SUQP_i == MPID_recvs.unexpected.tail) break;
			SUQP_prev = SUQP_i;
			SUQP_i = MPID_recvs.unexpected.queue[SUQP_i].next;
		}
		if (SUQP_p) {
			SUQP_rhandle =
			MPID_recvs.unexpected.queue[SUQP_i].ptr;
			/*  Take element "p" (index "i") out.  */
			SUQP_p->free = 1;
			MPID_recvs.unexpected.how_many--;
			if (!MPID_recvs.unexpected.how_many) {
				MPID_recvs.unexpected.head = 0;
				MPID_recvs.unexpected.tail = 0;
				return;
			}
			MPID_recvs.unexpected.queue[SUQP_prev].next =
				SUQP_p->next;
			if ((unsigned long)SUQP_i == MPID_recvs.unexpected.head)
				MPID_recvs.unexpected.head = SUQP_p->next;
			if ((unsigned long)SUQP_i == MPID_recvs.unexpected.tail)
				MPID_recvs.unexpected.tail = SUQP_prev;
			return;
		}
	}
	SUQP_rhandle = NULL;
	return;
}

void MPID_Atomic_Extract_unexpected_entry(
	int position, MPIR_RHANDLE **rhandleptr
)
{
	SUQP_position = position;
	gamma_atomic(do_MPID_Atomic_Extract_unexpected_entry);
	*rhandleptr = SUQP_rhandle;
}


void MPID_InitQueue()
{
	int i;

	DEBUG(printf("[%d] About to setup message queues\n",MPID_MyWorldRank);)
	/* Allocate queues */
	MPID_recvs.unexpected.queue = (MPID_QEL *)
		malloc(AG_MAXQLEN*sizeof(MPID_QEL));
	MPID_recvs.posted.queue = (MPID_QEL *)
		malloc(AG_MAXQLEN*sizeof(MPID_QEL));
	/* Pin-down queue-related data structures and functions */
	gamma_mlock_funct(
		do_MPID_Atomic_Search_unexpected_queue_and_post, 4096
	);
	gamma_mlock(&MPID_recvs, sizeof(MPID_recvs));
	gamma_mlock(&MPID_recvs.unexpected.how_many,
		sizeof(MPID_recvs.unexpected.how_many));
	gamma_mlock(&MPID_recvs.unexpected.head,
		sizeof(MPID_recvs.unexpected.head));
	gamma_mlock(&MPID_recvs.unexpected.tail,
		sizeof(MPID_recvs.unexpected.tail));
	gamma_mlock(MPID_recvs.unexpected.queue, AG_MAXQLEN*sizeof(MPID_QEL));
	gamma_mlock(&MPID_recvs.posted.how_many,
		sizeof(MPID_recvs.posted.how_many));
	gamma_mlock(&MPID_recvs.posted.head,
		sizeof(MPID_recvs.posted.head));
	gamma_mlock(&MPID_recvs.posted.tail,
		sizeof(MPID_recvs.posted.tail));
	gamma_mlock(MPID_recvs.posted.queue, AG_MAXQLEN*sizeof(MPID_QEL));
	gamma_mlock(&SUQP_src_lrank, sizeof(SUQP_src_lrank));
	gamma_mlock(&SUQP_tag, sizeof(SUQP_tag));
	gamma_mlock(&SUQP_context_id, sizeof(SUQP_context_id));
	gamma_mlock(&SUQP_flag, sizeof(SUQP_flag));
	gamma_mlock(&SUQP_position, sizeof(SUQP_position));
	gamma_mlock(&SUQP_len, sizeof(SUQP_len));
	gamma_mlock(&SUQP_i, sizeof(SUQP_i));
	gamma_mlock(&SUQP_prev, sizeof(SUQP_prev));
	gamma_mlock(&SUQP_buf, sizeof(SUQP_buf));
	gamma_mlock(&SUQP_request, sizeof(SUQP_request));
	gamma_mlock(&SUQP_rhandle, sizeof(SUQP_rhandle));
	gamma_mlock(&SUQP_p, sizeof(SUQP_p));
	/* Initialize queues */
	MPID_recvs.unexpected.how_many = 0;
	MPID_recvs.unexpected.head = 0;
	MPID_recvs.unexpected.tail = 0;
	MPID_recvs.posted.how_many = 0;
	MPID_recvs.posted.head = 0;
	MPID_recvs.posted.tail = 0;
	for (i=0;i<AG_MAXQLEN;i++) {
		MPID_recvs.unexpected.queue[i].free = 1;
		MPID_recvs.posted.queue[i].free = 1;
		MPID_recvs.posted.queue[i].rndv_pending = 0;
	}
	gamma_mlock(&AG_tempbuf_ptr, sizeof(AG_tempbuf_ptr));
}
